static char rcsid[] = "@(#)$Id: aliaslib.c,v 2.0 2007/08/25 11:51:59 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.0 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@posti.FMI.FI> 
 *                       (was hurtta+elm@ozone.FMI.FI)
 *           or  Kari Hurtta <elm@elmme-mailer.org>
 ******************************************************************************
 *  Based on Elm 2.4 src/aliaslib.c. That code was following copyright:
 *
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** Library of functions dealing with the alias system...

 **/

#include "def_elm.h"
#include "s_elm.h"

DEBUG_VAR(Debug,__FILE__,"alias");

extern int current_mail_message;

static int add_ P_((struct addr_list *list, 
		    const char *ADR, 
		    const struct string *FN,
		    const struct string *COMMENT,
		    int group_num));
static int add_(list,ADR,FN,COMMENT,group_num)
     struct addr_list *list;
     const char *ADR;
     const struct string *FN;
     const struct string *COMMENT;
     int group_num;
{
    int x;

    struct address * address =  new_address(ADR,FN,COMMENT);

    x = add_address_to_list(list,address,group_num);

    free_address(&address);

    return x;
}

static int do_get_alias_l P_((
			      char *name,
			      struct addr_list *result_list,
			      int mailing,
			      int sysalias,
			      int depth,
			      int *too_longp,
			      struct mailer_info  *mailer_info,
			      struct AliasView *aview,
			      int group_num
			      ));
/*
 * Expand the comma-delimited group of names in "group", storing the result
 * in "buffer".  Returns TRUE if expansion occurs OK, else FALSE in the
 * event of errors.
 */
static int do_expand_group_l P_((
				 char *group,
				 struct addr_list *result_list,
				 int sysalias,
				 int depth,
				 int *too_longp,
				 struct mailer_info  *mailer_info,
				 struct AliasView *aview,
				 struct string * group_phrase
				 ));

static int do_expand_group_l(group, result_list, sysalias, depth, too_longp,
			     mailer_info, aview, group_phrase)
     char *group;	     /* group list to expand			*/
     struct addr_list   *result_list;  /* place to store result of expansion	*/
     int sysalias;	/* TRUE to suppress checks of the user's aliases*/
     int depth;	/* nesting depth					*/
     int *too_longp;	/* error code if expansion overflows buffer	*/
     struct mailer_info  *mailer_info;
     struct AliasView *aview;
     struct string * group_phrase;
{
    char ** tokenized = rfc822_tokenize(group);
    char **ptr, **next = NULL;
    int result = 0;
    int group_num = -1;

    if (group_phrase && string_len(group_phrase) > 0)
	group_num =  add_group_to_list(result_list,group_phrase);
    

    remove_space_tokenized(tokenized);  /* removes spaces and comments */
    
    for (ptr = tokenized; *ptr; ptr = next) {
	char * name = *ptr;
	int q_seen = 0;
	int q = 0;
	
	for (next = ptr; *next; next++) {
	    if (',' == (*next)[0] && !q)
		break;
	    else if ('<' == (*next)[0]) {
		q++;
		q_seen++;
	    } else if ('>' == (*next)[0])
		q--;
	}
	
	if (ptr+1 == next) {
	    struct addr_verify_result verify_result;
	    enum mailer_errcode errcode = MAILER_NOT_AVAIL;

	    /* see if this name is really an alias */
	    if (do_get_alias_l(name,result_list,TRUE,sysalias,depth,too_longp,
			       mailer_info, aview,group_num))
		continue;
	    
	    if ( *too_longp ) {
		result = 0;
		goto fail;
	    }

	    /* verify it is a valid address */
	    if (mailer_info &&
		verify_mailer_addr(mailer_info,name,
			       &verify_result,&errcode)) {

		DPRINT(Debug,7,
		       (&Debug, 
			"Alias %s verified as address -- really %s\n",
			name,verify_result.addr));

		if (!verify_result.fullname)
		    verify_result.fullname = new_string(display_charset);
		if (!verify_result.comment)
		    verify_result.comment = new_string(display_charset);


		/* Max alias expansion limit? */
		add_(result_list,
		     verify_result.addr,
		     verify_result.fullname,
		     verify_result.comment,
		     group_num);      
		
		free_string(& verify_result.fullname);
		free_string(& verify_result.comment);
		free(verify_result.addr); verify_result.addr = NULL;

	    } else {
		struct string * fn = NULL;
		struct string * cm = NULL;

		cm = new_string(display_charset);

		fn = new_string(display_charset);

		/* Max alias expansion limit? */
		add_(result_list,name,fn,cm,group_num);      

		free_string(&fn);
		free_string(&cm);
	    }

	} else {  /* must have address .... */
	    char * buffer1 = NULL;
	    char ** scanner = NULL;
	    
	    for (scanner = ptr; scanner < next; scanner++) {
		if (buffer1 && q_seen && !q)
		    buffer1 = strmcat(buffer1," ");  /* Regenerate spaces for phrase */
		if ('<' == (*scanner)[0]) 
		    q++;
		else if ('>' == (*scanner)[0])
		    q--;
		buffer1 = strmcat(buffer1,*scanner);
	    }
	    
	    if (buffer1) {
		struct addr_list * address_list =
		    parse_header_address(NULL,buffer1, 
					 /* If user pastes encoded words from 
					  * somewhere decode them also.
					  */
					 is_rfc1522(buffer1),
					 display_charset,NULL);

		int idx, addr_item_count = addr_list_item_count(address_list);


		for (idx = 0; idx < addr_item_count; idx++) {
		    int group = -1;
		    const struct address * address = 
			addr_list_get_item(address_list,idx,&group);
		    int g = group_num;

		    if (-1 == g && group >= 0) {
			const struct string *gname = 
			    addr_list_get_group(address_list,group);
			int i;
			int l1 = addr_list_group_count(result_list);
			
			for (i = 0; i < l1; i++) {
			    const struct string *gname1 = 
				addr_list_get_group(result_list,i);
			    if (0 == string_cmp(gname,gname1,-99))
				 g = i;
			}	

			if (-1 == g)
			    g = add_group_to_list(result_list,gname);
		    }

		    /* Max alias expansion limit? */
		    add_address_to_list(result_list,address,g);

		}
		if (address_list)
		    free_addr_list(&address_list);
		
		free(buffer1);
	    }      
	}
	
	if (*next &&
	    ',' == (*next)[0])
	    next++;
    }

    result = 1;
 fail:

    if (tokenized)
	free_rfc822tokenized(tokenized);
    
    return result;
}


static int do_get_alias_l(name, result_list, mailing, sysalias, depth, too_longp,
			  mailer_info, aview, group_num)
     char *name;	    /* name to expand as an alias		     */
     struct addr_list *result_list; /* place to store result of expansion    */
     int mailing;	   /* TRUE to fully expand group names & recursive aliases	*/
     int sysalias;	   /* TRUE to suppress checks of the user's aliases  */
     int depth;	           /* recursion depth - initially call at depth=0    */
     int *too_longp;	   /* error code if expansion overflows buffer	     */
     struct mailer_info  *mailer_info;
     struct AliasView *aview;
     int group_num;

{
    struct alias_rec *match = NULL;
    char * alias_address = NULL;
    struct string * alias_fullname = NULL;
    struct string * alias_comment = NULL;
    
    struct addr_list * address_list = NULL;
    int addr_item_count;
    int loc;
    
    /* update the recursion depth counter */
    ++depth;
    
    DPRINT(Debug,5,
	   (&Debug, "%*s->attempting alias expansion on \"%s\"\n",
	    (depth*2), "", name));
    
    /* The next two blocks could be merged somewhat */
    /* check for a user alias, unless in the midst of sys alias expansion */
    if ( !sysalias &&
	 (loc = find_alias(name, USER, aview)) >= 0)
	match = give_alias(aview,loc);

    /* check for a system alias */
    else if ( (loc = find_alias(name, SYSTEM, aview)) >= 0 ) {
	match = give_alias(aview,loc);
	sysalias = TRUE;
    }
    
    /* nope...this name wasn't an alias */
    else
	return FALSE;
    
    if (!match)
	return FALSE;

    alias_address = match->address;

    
    if (alias_group_phrase || 
	0 != (match->type & PERSON )) {
	/* If user pastes encoded words from 
	 *	somewhere decode them also.
	 */	
	alias_fullname = hdr_to_string(HDR_TEXT,match->name,
				       display_charset,
				       is_rfc1522(match->name));
    }

    DPRINT(Debug,5,
	   (&Debug, "%*s  ->expanded alias to \"%s\"\n",
	    (depth*2), "", alias_address));

    if (!alias_fullname)
	alias_fullname = new_string(display_charset);
    if (!alias_comment)
	alias_comment = new_string(display_charset);


    if (alias_fullname) {
	DPRINT(Debug,5,
	       (&Debug, "%*s  --> fullname \"%S\"\n",
		(depth*2), "", alias_fullname));
    }

    
    /* check for an exact match */
    if (0 == strcmp(name,alias_address)) {
	add_(address_list,alias_address,alias_fullname,alias_comment,group_num);
	
	free_string(&alias_fullname);
	free_string(&alias_comment);

	/* Max alias expansion limit? */
	return TRUE;
    }
  
    /* see if we are stuck in a loop */
    if ( depth > 12 ) {
	DPRINT(Debug,2,
	       (&Debug,
		"alias expansion loop detected at \"%s\" - bailing out\n", 
		name));
	lib_error(CATGETS(elm_msg_cat, ElmSet, ElmErrorExpanding,
			  "Error expanding \"%s\" - probable alias definition loop."),
		  name);
	
	free_string(&alias_fullname);
	free_string(&alias_comment);
	
	return FALSE;
    }
    
    /* see if the alias equivalence is a group name */
    if ( mailing && match->type & GROUP ) {
	int r;

	r =  do_expand_group_l(alias_address,result_list,
			       sysalias,depth,too_longp,
			       mailer_info, aview, alias_fullname
			       );

	free_string(&alias_fullname);
	free_string(&alias_comment);


	return r;
    }
    
    /* see if the alias equivalence is an email address,
     * but not list of addresses. If address starts with @
     * then it is source-route address and not list of addresses
     */
    if (!(match->type & GROUP) &&
	('@' == alias_address[0] || qstrpbrk(alias_address,",") == NULL) &&
	qstrpbrk(alias_address,"!@:") != NULL ) {

	add_(result_list,alias_address,alias_fullname,alias_comment,group_num);
    
	free_string(&alias_fullname);
	free_string(&alias_comment);

	/* Max alias expansion limit? */
	return TRUE;
    }
    
    /* see if the alias equivalence is itself an alias */
    if ( mailing && qstrpbrk(alias_address,",!@:") == NULL &&
	 do_get_alias_l(alias_address,result_list,TRUE,sysalias,depth,
			too_longp,mailer_info, aview, group_num))
	return TRUE;
    
    /* the alias equivalence must just be a local address  or 
       list of addresses */
    
     address_list = parse_header_address(NULL,alias_address, 
					 /* If user pastes encoded words from 
					  *	somewhere decode them also.
					  */
					 is_rfc1522(alias_address),
					 display_charset,
					 NULL);

     addr_item_count = addr_list_item_count(address_list);
    
    /* only one address -- take fullname from alias */
     if (1 == addr_item_count) {
	 int group = -1;

	 const struct address * address = 
	     addr_list_get_item(address_list,0,&group);
	 int g = group_num;

	 const char          * addr     = address_get_ascii_addr(address);
	 const struct string * fullname = address_get_phrase(address);
	 const struct string * comment  = address_get_comment(address);

    
	 if (-1 == g && group >= 0) {
	     const struct string *gname = 
		 addr_list_get_group(address_list,group);
	     int i;
	     int l1 = addr_list_group_count(result_list);
	     
	     for ( i = 0; i < l1; i++) {
		 const struct string *gname1 = 
		     addr_list_get_group(result_list,i);
		 if (0 == string_cmp(gname,gname1,-99))
		     g = i;
	     }	
	     
	     if (-1 == g)
		 g = add_group_to_list(result_list,gname);
	 }
	 
	 if (addr)
	     add_(result_list,addr,alias_fullname,alias_comment,g);
	 else 
	     goto handle_empty;
     } else {

	 int idx;
       	
     handle_empty:
	/* List of addresses */

	 for (idx = 0; idx < addr_item_count; idx++) {
	     int group = -1;
	     const struct address * address = 
		 addr_list_get_item(address_list,idx,&group);
	     int g = group_num;

	     if (-1 == g && group >= 0) {
		 const struct string *gname = 
		     addr_list_get_group(address_list,group);
		 int i;
		 int l1 = addr_list_group_count(result_list);
		 
		 for ( i = 0; i < l1; i++) {
		     const struct string *gname1 = 
			 addr_list_get_group(result_list,i);
		     if (0 == string_cmp(gname,gname1,-99))
			 g = i;
		 }	
		 
		 if (-1 == g)
		     g = add_group_to_list(result_list,gname);
	     }
	     
	     /* Max alias expansion limit? */
	     add_address_to_list(result_list,address,g);	     
	 }
	 
	 if (address_list)
	     free_addr_list(&address_list);
    } 

    free_string(&alias_fullname);
    free_string(&alias_comment);

    return TRUE;
}

struct addr_list *get_alias_address_l (name, mailing, too_longp,mailer_info,
				       aview)
     char *name; /* name to expand as an alias */
     int mailing; /* TRUE to fully expand group names & recursive aliases */
     int *too_longp; /* error code if expansion overflows buffer          */
     struct mailer_info  *mailer_info;
     struct AliasView *aview;
{   
    struct addr_list * result = new_empty_addr_list(1);
    
    /*
     *	Reopens files iff changed since last read
     */
    open_alias_files(aview, NULL);
    /*
     *	If name is an alias then return its expansion
     */
    
    if (do_get_alias_l(name,result,mailing,FALSE,0,too_longp,
		       mailer_info, aview, -1)) {

	int alen  = addr_list_item_count(result);
	int glen = addr_list_group_count(result);
	int i;

	DPRINT(Debug,2,
	       (&Debug,
		"get_alias_address_l(%s) expands %d addresses (with %d groups).\n",
		name,alen,glen));

	
	for (i = 0; i < glen;  i++) {
	    const struct string * G = addr_list_get_group(result,i);
	    
	    DPRINT(Debug,6,(&Debug,"get_alias_address_l: group [%d]: group=%S\n",
			    i,G));
	}
	
	for (i = 0; i < alen;  i++) {
	    
	    int group = -1;
	    const struct address * address = 
		addr_list_get_item(result,i,&group);
	    
	    const char          * addr     = address_get_ascii_addr(address);
	    const struct string * fullname = address_get_phrase(address);
	    const struct string * comment  = address_get_comment(address);
	    
	    DPRINT(Debug,6,(&Debug,"get_alias_address_l: addr [%d]: addr=%s",
			    i,addr ? addr : "<NULL>"));
	    
	    if (fullname) {
		const char * A = get_string_MIME_name(fullname);
		const char * B = get_string_lang(fullname);	    
		
		DPRINT(Debug,6,(&Debug,", fullname=%S",
				fullname));
	    }
	    
	    if (comment) {
		const char * A = get_string_MIME_name(comment);
		const char * B = get_string_lang(comment);	    
		
		DPRINT(Debug,6,(&Debug,", comment=%S",comment));
	    }
	    
	    DPRINT(Debug,6,(&Debug,", group=%d\n",group));
	}

    } else {

	/*
	 *  Nope...not an alias (or it was too long to expand)
	 */
	DPRINT(Debug,2,
	       (&Debug,
		"Could not expand alias in get_alias_address_l(%s)%s\n",
		name, *too_longp ? "\t...alias buffer overflowed." : ""));

	if (result)
	    free_addr_list(&result);
    }
        
    return(result);    
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
